#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import numbers
import weakref
import warnings

from copy import deepcopy
from types import MappingProxyType
from typing import (
    List,
    Union,
    Any,
    Dict,
    Optional,
    TYPE_CHECKING,
    Callable,
    Tuple,
    TypeVar,
)

from easyCore import borg, ureg, np, pint
from easyCore.Utils.classTools import addProp
from easyCore.Utils.Exceptions import CoreSetException
from easyCore.Utils.UndoRedo import property_stack_deco
from easyCore.Utils.json import MSONable
from easyCore.Fitting.Constraints import SelfConstraint

if TYPE_CHECKING:
    from easyCore.Utils.typing import C

Q_ = ureg.Quantity
M_ = ureg.Measurement


class Descriptor(MSONable):
    """
    This is the base of all variable descriptions for models. It contains all information to describe a single
    unique property of an object. This description includes a name and value as well as optionally a unit, description
    and url (for reference material). Also implemented is a callback so that the value can be read/set from a linked
    library object.

    A `Descriptor` is typically something which describes part of a model and is non-fittable and generally changes the
    state of an object.
    """

    _constructor = Q_
    _borg = borg

    def __init__(
        self,
        name: str,
        value: Any,
        units: Optional[Union[str, ureg.Unit]] = None,
        description: Optional[str] = "",
        url: Optional[str] = "",
        display_name: Optional[str] = None,
        callback: Optional[property] = property(),
        enabled: Optional[bool] = True,
        parent: Optional[Union[Any, None]] = None,
    ):  # noqa: S107
        """
        This is the base of all variable descriptions for models. It contains all information to describe a single
        unique property of an object. This description includes a name and value as well as optionally a unit,
        description and url (for reference material). Also implemented is a callback so that the value can be read/set
        from a linked library object.

        A `Descriptor` is typically something which describes part of a model and is non-fittable and generally changes
        the state of an object.

        Units are provided by pint: https://github.com/hgrecco/pint

        :param name: Name of this object
        :param value: Value of this object
        :param units: This object can have a physical unit associated with it
        :param description: A brief summary of what this object is
        :param url: Lookup url for documentation/information
        :param callback: The property which says how the object is linked to another one
        :param parent: The object which is the parent to this one

        .. code-block:: python

             from easyCore.Objects.Base import Descriptor
             # Describe a color by text
             color_text = Descriptor('fav_colour', 'red')
             # Describe a color by RGB
             color_num = Descriptor('fav_colour', [1, 0, 0])

        .. note:: Undo/Redo functionality is implemented for the attributes `value`, `unit` and `display name`.
        """
        if not hasattr(self, "_args"):
            self._args = {"value": None, "units": ""}

        # Let the collective know we've been assimilated
        self._borg.map.add_vertex(self, obj_type="created")
        # Make the connection between self and parent
        if parent is not None:
            self._borg.map.add_edge(parent, self)

        self.name: str = name
        # Attach units if necessary
        if isinstance(units, ureg.Unit):
            self._units = ureg.Quantity(1, units=deepcopy(units))
        elif isinstance(units, (str, type(None))):
            self._units = ureg.parse_expression(units)
        else:
            raise AttributeError
        # Clunky method of keeping self.value up to date
        self._type = type(value)
        self.__isBooleanValue = isinstance(value, bool)
        if self.__isBooleanValue:
            value = int(value)
        self._args["value"] = value
        self._args["units"] = str(self.unit)
        self._value = self.__class__._constructor(**self._args)

        self._enabled = enabled

        self.description: str = description
        self._display_name: str = display_name
        self.url: str = url
        if callback is None:
            callback = property()
        self._callback: property = callback
        self.user_data: dict = {}

        finalizer = None
        if self._callback.fdel is not None:
            weakref.finalize(self, self._callback.fdel)
        self._finalizer = finalizer

    def __reduce__(self):
        """
        Make the class picklable. Due to the nature of the dynamic class definitions special measures need to be taken.

        :return: Tuple consisting of how to make the object
        :rtype: tuple
        """
        state = self.as_dict()
        cls = self.__class__
        if hasattr(self, "__old_class__"):
            cls = self.__old_class__
        return cls.from_dict, (state,)

    @property
    def display_name(self) -> str:
        """
        Get a pretty display name.

        :return: The pretty display name.
        """
        # TODO This might be better implementing fancy f-strings where we can return html,latex, markdown etc
        display_name = self._display_name
        if display_name is None:
            display_name = self.name
        return display_name

    @display_name.setter
    @property_stack_deco
    def display_name(self, name_str: str):
        """
        Set the pretty display name.

        :param name_str: Pretty display name of the object.
        :return: None
        """
        self._display_name = name_str

    @property
    def unit(self) -> pint.UnitRegistry:
        """
        Get the unit associated with the object.

        :return: Unit associated with self in `pint` form.
        """
        return self._units.units

    @unit.setter
    @property_stack_deco
    def unit(self, unit_str: str):
        """
        Set the unit to a new one.

        :param unit_str: String representation of the unit required. i.e `m/s`
        :return: None
        """
        if not isinstance(unit_str, str):
            unit_str = str(unit_str)
        new_unit = ureg.parse_expression(unit_str)
        self._units = new_unit
        self._args["units"] = str(new_unit)
        self._value = self.__class__._constructor(**self._args)

    @property
    def value(self) -> Any:
        """
        Get the value of self as a pint. This is should be usable for most cases. If a pint
        is not acceptable then the raw value can be obtained through `obj.raw_value`.

        :return: Value of self with unit.
        """
        # Cached property? Should reference callback.
        # Also should reference for undo/redo
        if self._callback.fget is not None:
            try:
                value = self._callback.fget()
                if value != self._value:
                    self.__deepValueSetter(value)
            except Exception as e:
                raise ValueError(f"Unable to return value:\n{e}")
        r_value = self._value
        if self.__isBooleanValue:
            r_value = bool(r_value)
        return r_value

    def __deepValueSetter(self, value: Any):
        """
        Set the value of self. This creates a pint with a unit.

        :param value: New value of self
        :return: None
        """
        # TODO there should be a callback to the collective, logging this as a return(if from a non `easyCore` class)
        if hasattr(value, "magnitude"):
            value = value.magnitude
            if hasattr(value, "nominal_value"):
                value = value.nominal_value
        self._type = type(value)
        self.__isBooleanValue = isinstance(value, bool)
        if self.__isBooleanValue:
            value = int(value)
        self._args["value"] = value
        self._value = self.__class__._constructor(**self._args)

    @value.setter
    @property_stack_deco
    def value(self, value: Any):
        """
        Set the value of self. This creates a pint with a unit.

        :param value: New value of self
        :return: None
        """
        if not self.enabled:
            if borg.debug:
                raise CoreSetException(f"{str(self)} is not enabled.")
            return
        self.__deepValueSetter(value)
        if self._callback.fset is not None:
            try:
                self._callback.fset(value)
            except Exception as e:
                raise CoreSetException(e)

    @property
    def raw_value(self) -> Any:
        """
        Return the raw value of self without a unit.

        :return: The raw value of self
        """
        value = self._value
        if hasattr(value, "magnitude"):
            value = value.magnitude
            if hasattr(value, "nominal_value"):
                value = value.nominal_value
        if self.__isBooleanValue:
            value = bool(value)
        return value

    @property
    def enabled(self) -> bool:
        """
        Logical property to see if the objects value can be directly set.

        :return: Can the objects value be set
        """
        return self._enabled

    @enabled.setter
    @property_stack_deco
    def enabled(self, value: bool):
        """
        Enable and disable the direct setting of an objects value field.

        :param value: True - objects value can be set, False - the opposite
        """
        self._enabled = value

    def convert_unit(self, unit_str: str):
        """
        Convert the value from one unit system to another. You will should use
        `compatible_units` to see if your new unit is compatible.

        :param unit_str: New unit in string form
        """
        new_unit = ureg.parse_expression(unit_str)
        self._value = self._value.to(new_unit)
        self._units = new_unit
        self._args["value"] = self.raw_value
        self._args["units"] = str(self.unit)

    # @cached_property
    @property
    def compatible_units(self) -> List[str]:
        """
        Returns all possible units for which the current unit can be converted.

        :return: Possible conversion units
        """
        return [str(u) for u in self.unit.compatible_units()]

    def __repr__(self):
        """Return printable representation of a Descriptor/Parameter object."""
        class_name = self.__class__.__name__
        obj_name = self.name
        if self.__isBooleanValue:
            obj_value = self.raw_value
        else:
            obj_value = self._value.magnitude
        if isinstance(obj_value, float):
            obj_value = "{:0.04f}".format(obj_value)
        obj_units = ""
        if not self.unit.dimensionless:
            obj_units = " {:~P}".format(self.unit)
        out_str = f"<{class_name} '{obj_name}': {obj_value}{obj_units}>"
        return out_str

    def as_dict(self, skip: List[str] = None) -> Dict[str, str]:
        """
        Convert ones self into a serialized form.

        :return: dictionary of ones self
        """
        if skip is None:
            skip = []
        super_dict = super().as_dict(skip=skip + ["parent", "callback", "_finalizer"])
        super_dict["value"] = self.raw_value
        super_dict["units"] = self._args["units"]
        # Attach the id. This might be useful in connected applications.
        # Note that it is converted to int and then str because javascript....
        super_dict["@id"] = str(self._borg.map.convert_id(self).int)
        return super_dict

    def to_obj_type(self, data_type: Parameter, *kwargs):
        """
        Convert between a `Parameter` and a `Descriptor`.

        :param data_type: class constructor of what we want to be
        :param kwargs: Additional keyword/value pairs for conversion
        :return: self as a new type
        """
        pickled_obj = self.as_dict()
        pickled_obj.update(kwargs)
        return data_type.from_dict(pickled_obj)

    def __copy__(self):
        return self.__class__.from_dict(self.as_dict())


V = TypeVar("V", bound=Descriptor)


class ComboDescriptor(Descriptor):
    """
    This class is an extension of a ``easyCore.Object.Base.Descriptor``. This class has a selection of values which can
    be checked against. For example, combo box styling.
    """

    def __init__(self, *args, available_options: list = None, **kwargs):
        super(ComboDescriptor, self).__init__(*args, **kwargs)
        if available_options is None:
            available_options = []
        self._available_options = available_options

        # We have initialized from the Descriptor class where value has it's own undo/redo decorator
        # This needs to be bypassed to use the Parameter undo/redo stack
        fun = self.__class__.value.fset
        if hasattr(fun, "func"):
            fun = getattr(fun, "func")
        self.__previous_set: Callable[
            [V, Union[numbers.Number, np.ndarray]], Union[numbers.Number, np.ndarray]
        ] = fun

        # Monkey patch the unit and the value to take into account the new max/min situation
        addProp(
            self,
            "value",
            fget=self.__class__.value.fget,
            fset=self.__class__._property_value.fset,
            fdel=self.__class__.value.fdel,
        )

    @property
    def _property_value(self) -> Union[numbers.Number, np.ndarray]:
        return self.value

    @_property_value.setter
    @property_stack_deco
    def _property_value(self, set_value: Union[numbers.Number, np.ndarray, Q_]):
        """
        Verify value against constraints. This hasn't really been implemented as fitting is tricky.

        :param set_value: value to be verified
        :return: new value from constraint
        """
        if isinstance(set_value, Q_):
            set_value = set_value.magnitude
        # Save the old state and create the new state
        old_value = self._value
        state = self._borg.stack.enabled
        if state:
            self._borg.stack.force_state(False)
        try:
            new_value = old_value
            if set_value in self.available_options:
                new_value = set_value
        finally:
            self._borg.stack.force_state(state)

        # Restore to the old state
        self.__previous_set(self, new_value)

    @property
    def available_options(self) -> List[Union[numbers.Number, np.ndarray, Q_]]:
        return self._available_options

    @available_options.setter
    @property_stack_deco
    def available_options(
        self, available_options: List[Union[numbers.Number, np.ndarray, Q_]]
    ) -> None:
        self._available_options = available_options

    def as_dict(self, **kwargs) -> Dict[str, Any]:
        import json

        d = super().as_dict(**kwargs)
        d["name"] = self.name
        d["available_options"] = json.dumps(self.available_options)
        return d


class Parameter(Descriptor):
    """
    This class is an extension of a ``easyCore.Object.Base.Descriptor``. Where the descriptor was for static objects,
    a `Parameter` is for dynamic objects. A parameter has the ability to be used in fitting and
    has additional fields to facilitate this.
    """

    _constructor = M_

    def __init__(
        self,
        name: str,
        value: Union[numbers.Number, np.ndarray],
        error: Optional[Union[numbers.Number, np.ndarray]] = 0.0,
        min: Optional[numbers.Number] = -np.Inf,
        max: Optional[numbers.Number] = np.Inf,
        fixed: Optional[bool] = False,
        **kwargs,
    ):
        """
        This class is an extension of a ``easyCore.Object.Base.Descriptor``. Where the descriptor was for static
        objects, a `Parameter` is for dynamic objects. A parameter has the ability to be used in fitting and has
        additional fields to facilitate this.

        :param name: Name of this obj
        :param value: Value of this object
        :param error: Error associated as sigma for this parameter
        :param min: Minimum value for fitting
        :param max: Maximum value for fitting
        :param fixed: Should this parameter vary when fitting?
        :param kwargs: Key word arguments for the `Descriptor` class.

        .. code-block:: python

             from easyCore.Objects.Base import Parameter
             # Describe a phase
             phase_basic = Parameter('phase', 3)
             # Describe a phase with a unit
             phase_unit = Parameter('phase', 3, units,='rad/s')

        .. note::
            Undo/Redo functionality is implemented for the attributes `value`, `error`, `min`, `max`, `fixed`
        """
        # Set the error
        self._args = {"value": value, "units": "", "error": error}

        if not isinstance(value, numbers.Number):
            raise ValueError("In a parameter the `value` must be numeric")
        if value < min:
            raise ValueError("`value` can not be less than `min`")
        if value > max:
            raise ValueError("`value` can not be greater than `max`")
        if error < 0:
            raise ValueError("Standard deviation `error` must be positive")

        super().__init__(name, value, **kwargs)
        self._args["units"] = str(self.unit)

        # Warnings if we are given a boolean
        if self._type == bool:
            warnings.warn(
                "Boolean values are not officially supported in Parameter. Use a Descriptor instead",
                UserWarning,
            )

        # Create additional fitting elements
        self._min: numbers.Number = min
        self._max: numbers.Number = max
        self._fixed: bool = fixed
        self.initial_value = self.value
        self._constraints: dict = {
            "user": {},
            "builtin": {
                "min": SelfConstraint(self, ">=", "_min"),
                "max": SelfConstraint(self, "<=", "_max"),
            },
            "virtual": {},
        }
        # This is for the serialization. Otherwise we wouldn't catch the values given to `super()`
        self._kwargs = kwargs

        # We have initialized from the Descriptor class where value has it's own undo/redo decorator
        # This needs to be bypassed to use the Parameter undo/redo stack
        fun = self.__class__.value.fset
        if hasattr(fun, "func"):
            fun = getattr(fun, "func")
        self.__previous_set: Callable[
            [V, Union[numbers.Number, np.ndarray]],
            Union[numbers.Number, np.ndarray],
        ] = fun

        # Monkey patch the unit and the value to take into account the new max/min situation
        addProp(
            self,
            "value",
            fget=self.__class__.value.fget,
            fset=self.__class__._property_value.fset,
            fdel=self.__class__.value.fdel,
        )

    @property
    def _property_value(self) -> Union[numbers.Number, np.ndarray, M_]:
        return self.value

    @_property_value.setter
    @property_stack_deco
    def _property_value(self, set_value: Union[numbers.Number, np.ndarray, M_]) -> None:
        """
        Verify value against constraints. This hasn't really been implemented as fitting is tricky.

        :param set_value: value to be verified
        :return: new value from constraint
        """
        if isinstance(set_value, M_):
            set_value = set_value.magnitude.nominal_value
        # Save the old state and create the new state
        old_value = self._value
        self._value = self.__class__._constructor(
            value=set_value, units=self._args["units"], error=self._args["error"]
        )

        # First run the built in constraints. i.e. min/max
        constraint_type: MappingProxyType[str, C] = self.builtin_constraints
        new_value = self.__constraint_runner(constraint_type, set_value)
        # Then run any user constraints.
        constraint_type: dict = self.user_constraints
        state = self._borg.stack.enabled
        if state:
            self._borg.stack.force_state(False)
        try:
            new_value = self.__constraint_runner(constraint_type, new_value)
        finally:
            self._borg.stack.force_state(state)

        # And finally update any virtual constraints
        constraint_type: dict = self._constraints["virtual"]
        _ = self.__constraint_runner(constraint_type, new_value)

        # Restore to the old state
        self._value = old_value
        self.__previous_set(self, new_value)

    def convert_unit(self, new_unit: str):  # noqa: S1144
        """
        Perform unit conversion. The value, max and min can change on unit change.

        :param new_unit: new unit
        :return: None
        """
        old_unit = str(self._args["units"])
        super().convert_unit(new_unit)
        # Deal with min/max. Error is auto corrected
        if not self.value.unitless and old_unit != "dimensionless":
            self._min = Q_(self.min, old_unit).to(self._units).magnitude
            self._max = Q_(self.max, old_unit).to(self._units).magnitude
        # Log the new converted error
        self._args["error"] = self.value.error.magnitude

    @property
    def min(self) -> numbers.Number:
        """
        Get the minimum value for fitting.

        :return: minimum value
        """
        return self._min

    @min.setter
    @property_stack_deco
    def min(self, value: numbers.Number):
        """
        Set the minimum value for fitting.
        - implements undo/redo functionality.

        :param value: new minimum value
        :return: None
        """
        if value <= self.raw_value:
            self._min = value
        else:
            raise ValueError(
                f"The current set value ({self.raw_value}) is less than the desired min value ({value})."
            )

    @property
    def max(self) -> numbers.Number:
        """
        Get the maximum value for fitting.

        :return: maximum value
        """
        return self._max

    @max.setter
    @property_stack_deco
    def max(self, value: numbers.Number):
        """
        Get the maximum value for fitting.
        - implements undo/redo functionality.

        :param value: new maximum value
        :return: None
        """
        if value >= self.raw_value:
            self._max = value
        else:
            raise ValueError(
                f"The current set value ({self.raw_value}) is greater than the desired max value ({value})."
            )

    @property
    def fixed(self) -> bool:
        """
        Can the parameter vary while fitting?

        :return: True = fixed, False = can vary
        :rtype: bool
        """
        return self._fixed

    @fixed.setter
    @property_stack_deco
    def fixed(self, value: bool):
        """
        Change the parameter vary while fitting state.
        - implements undo/redo functionality.

        :param value: True = fixed, False = can vary
        :return: None
        """
        if not self.enabled:
            if self._borg.stack.enabled:
                self._borg.stack.pop()
            if borg.debug:
                raise CoreSetException(f"{str(self)} is not enabled.")
            return
        # TODO Should we try and cast value to bool rather than throw ValueError?
        if not isinstance(value, bool):
            raise ValueError
        self._fixed = value

    @property
    def error(self) -> float:
        """
        The error associated with the parameter.

        :return: Error associated with parameter
        """
        return float(self._value.error.magnitude)

    @error.setter
    @property_stack_deco
    def error(self, value: float):
        """
        Set the error associated with the parameter.
        - implements undo/redo functionality.

        :param value: New error value
        :return: None
        """
        if value < 0:
            raise ValueError
        self._args["error"] = value
        self._value = self.__class__._constructor(**self._args)

    def __repr__(self) -> str:
        """
        Return printable representation of a Parameter object.
        """
        super_str = super().__repr__()
        super_str = super_str[:-1]
        s = []
        if self.fixed:
            super_str += " (fixed)"
        s.append(super_str)
        s.append("bounds=[%s:%s]" % (repr(self.min), repr(self.max)))
        return "%s>" % ", ".join(s)

    def __float__(self) -> float:
        return float(self.raw_value)

    def as_dict(self, skip: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Include enabled in the dict output as it's unfortunately skipped

        :param skip: Which items to skip when serializing
        :return: Serialized dictionary
        """
        new_dict = super(Parameter, self).as_dict(skip=skip)
        new_dict["enabled"] = self.enabled
        return new_dict

    @property
    def builtin_constraints(self) -> MappingProxyType[str, C]:
        """
        Get the built in constrains of the object. Typically these are the min/max

        :return: Dictionary of constraints which are built into the system
        """
        return MappingProxyType(self._constraints["builtin"])

    @property
    def user_constraints(self) -> Dict[str, C]:
        """
        Get the user specified constrains of the object.

        :return: Dictionary of constraints which are user supplied
        """
        return self._constraints["user"]

    @user_constraints.setter
    def user_constraints(self, constraints_dict: Dict[str, C]) -> None:
        self._constraints["user"] = constraints_dict

    def _quick_set(
        self,
        set_value: float,
        run_builtin_constraints: bool = False,
        run_user_constraints: bool = False,
        run_virtual_constraints: bool = False,
    ) -> None:
        """
        This is a quick setter for the parameter. It bypasses all the checks and constraints,
        just setting the value and issuing the interface callbacks.

        WARNING: This is a dangerous function and should only be used when you know what you are doing.
        """
        # First run the built-in constraints. i.e. min/max
        if run_builtin_constraints:
            constraint_type: MappingProxyType = self.builtin_constraints
            set_value = self.__constraint_runner(constraint_type, set_value)
        # Then run any user constraints.
        if run_user_constraints:
            constraint_type: dict = self.user_constraints
            state = self._borg.stack.enabled
            if state:
                self._borg.stack.force_state(False)
            try:
                set_value = self.__constraint_runner(constraint_type, set_value)
            finally:
                self._borg.stack.force_state(state)
        if run_virtual_constraints:
            # And finally update any virtual constraints
            constraint_type: dict = self._constraints["virtual"]
            _ = self.__constraint_runner(constraint_type, set_value)

        # Finally set the value
        self._property_value._magnitude._nominal_value = set_value
        self._args["value"] = set_value
        if self._callback.fset is not None:
            self._callback.fset(set_value)

    def __constraint_runner(
        self,
        this_constraint_type: Union[dict, MappingProxyType[str, C]],
        newer_value: numbers.Number,
    ) -> float:
        for constraint in this_constraint_type.values():
            if constraint.external:
                constraint()
                continue
            this_new_value = constraint(no_set=True)
            if this_new_value != newer_value:
                if borg.debug:
                    print(f"Constraint `{constraint}` has been applied")
                self._value = self.__class__._constructor(
                    value=this_new_value,
                    units=self._args["units"],
                    error=self._args["error"],
                )
            newer_value = this_new_value
        return newer_value

    @property
    def bounds(self) -> Tuple[numbers.Number, numbers.Number]:
        """
        Get the bounds of the parameter.

        :return: Tuple of the parameters minimum and maximum values
        """
        return self._min, self._max

    @bounds.setter
    def bounds(
        self, new_bound: Union[Tuple[numbers.Number, numbers.Number], numbers.Number]
    ) -> None:
        """
        Set the bounds of the parameter. *This will also enable the parameter*.

        :param new_bound: New bounds. This can be a tuple of (min, max) or a single number (min).
        For changing the max use (None, max_value).
        """
        # Macro checking and opening for undo/redo
        close_macro = False
        if self._borg.stack.enabled:
            self._borg.stack.beginMacro("Setting bounds")
            close_macro = True
        # Have we only been given a single number (MIN)?
        if isinstance(new_bound, numbers.Number):
            self.min = new_bound
        # Have we been given a tuple?
        if isinstance(new_bound, tuple):
            new_min, new_max = new_bound
            # Are there any None values?
            if isinstance(new_min, numbers.Number):
                self.min = new_min
            if isinstance(new_max, numbers.Number):
                self.max = new_max
        # Enable the parameter if needed
        if not self.enabled:
            self.enabled = True
        # This parameter is now not fixed.
        self.fixed = False
        # Close the macro if we opened it
        if close_macro:
            self._borg.stack.endMacro()



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.2.2"

import numpy as np

from easyCore.Objects.Borg import Borg
import pint

default_fitting_engine = "lmfit"

ureg = pint.UnitRegistry()
borg = Borg()
borg.instantiate_stack()
borg.stack.enabled = False



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.0.1"


import functools

from easyCore import np
from easyCore.Objects.ObjectClasses import BaseObj, Parameter
from easyCore.Objects.Groups import BaseCollection

from typing import ClassVar, Optional, Iterable, Union


def designate_calc_fn(func):
    @functools.wraps(func)
    def wrapper(obj, *args, **kwargs):
        for name in list(obj.__annotations__.keys()):
            func.__globals__["_" + name] = getattr(obj, name).raw_value
        return func(obj, *args, **kwargs)

    return wrapper


class Polynomial(BaseObj):
    """
    A polynomial model.

    Parameters
    ----------
    name : str
        The name of the model.
    degree : int
        The degree of the polynomial.
    """

    coefficients: ClassVar[BaseCollection]

    def __init__(self, name: str = "polynomial",
                 coefficients: Optional[Union[Iterable[Union[float, Parameter]], BaseCollection]] = None):
        super(Polynomial, self).__init__(name, coefficients=BaseCollection("coefficients"))
        if coefficients is not None:
            if issubclass(type(coefficients), BaseCollection):
                self.coefficients = coefficients
            elif isinstance(coefficients, Iterable):
                for index, item in enumerate(coefficients):
                    if issubclass(type(item), Parameter):
                        self.coefficients.append(item)
                    elif isinstance(item, float):
                        self.coefficients.append(Parameter(name="c{}".format(index), value=item))
                    else:
                        raise TypeError("Coefficients must be floats or Parameters")
            else:
                raise TypeError("coefficients must be a list or a BaseCollection")

    def __call__(self, x: np.ndarray, *args, **kwargs) -> np.ndarray:
        return np.polyval([c.raw_value for c in self.coefficients], x)

    def __repr__(self):
        s = []
        if len(self.coefficients) >= 1:
            s += [f"{self.coefficients[0].raw_value}"]
            if len(self.coefficients) >= 2:
                s += [f"{self.coefficients[1].raw_value}x"]
                if len(self.coefficients) >= 3:
                    s += [
                        f"{c.raw_value}x^{i+2}"
                        for i, c in enumerate(self.coefficients[2:])
                        if c.raw_value != 0
                    ]
        s.reverse()
        s = " + ".join(s)
        return "Polynomial({}, {})".format(self.name, s)


class Line(BaseObj):

    m: ClassVar[Parameter]
    c: ClassVar[Parameter]

    def __init__(self, m: Optional[Union[Parameter, float]] = None,
                 c: Optional[Union[Parameter, float]] = None):
        super(Line, self).__init__("line",
                                   m=Parameter("m", 1.),
                                   c=Parameter("c", 0.))
        if m is not None:
            self.m = m
        if c is not None:
            self.c = c

    # @designate_calc_fn can be used to inject parameters into the calculation function. i.e. _m = m.raw_value
    def __call__(self, x: np.ndarray, *args, **kwargs) -> np.ndarray:
        return self.m.raw_value * x + self.c.raw_value

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.m, self.c)



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

from typing import List, Optional
from numbers import Number

from easyCore.Fitting.fitting_template import Union, Callable, \
    FittingTemplate, np, FitResults, NameConverter, FitError

# Import dfols specific objects
import dfols


class DFO(FittingTemplate):  # noqa: S101
    """
    This is a wrapper to Derivative free optimisation: https://numericalalgorithmsgroup.github.io/dfols/
    """

    property_type = Number
    name = 'DFO_LS'

    def __init__(self, obj, fit_function: Callable):
        """
        Initialize the fitting engine with a `BaseObj` and an arbitrary fitting function.

        :param obj: Object containing elements of the `Parameter` class
        :type obj: BaseObj
        :param fit_function: function that when called returns y values. 'x' must be the first
                            and only positional argument. Additional values can be supplied by
                            keyword/value pairs
        :type fit_function: Callable
        """
        super().__init__(obj, fit_function)
        self.p_0 = {}

    def make_model(self, pars: Optional[List] = None) -> Callable:
        """
        Generate a model from the supplied `fit_function` and parameters in the base object.
        Note that this makes a callable as it needs to be initialized with *x*, *y*, *weights*

        :return: Callable model which returns residuals
        :rtype: Callable
        """
        fit_func = self._generate_fit_function()

        def outer(obj):
            def make_func(x, y, weights):
                par = {}
                if not pars:
                    for name, item in obj._cached_pars.items():
                        par['p' + str(name)] = item.raw_value
                else:
                    for item in pars:
                        par['p' + str(NameConverter().get_key(item))] = item.raw_value

                def residuals(x0) -> np.ndarray:
                    for idx, par_name in enumerate(par.keys()):
                        par[par_name] = x0[idx]
                    return (y - fit_func(x, **par)) / weights

                setattr(residuals, 'x', x)
                setattr(residuals, 'y', y)
                return residuals

            return make_func

        return outer(self)

    def _generate_fit_function(self) -> Callable:
        """
        Using the user supplied `fit_function`, wrap it in such a way we can update `Parameter` on
        iterations.

        :return: a fit function which is compatible with bumps models
        :rtype: Callable
        """
        # Original fit function
        func = self._original_fit_function
        # Get a list of `Parameters`
        self._cached_pars = {}
        for parameter in self._object.get_fit_parameters():
            self._cached_pars[NameConverter().get_key(parameter)] = parameter

        # Make a new fit function
        def fit_function(x: np.ndarray, **kwargs):
            """
            Wrapped fit function which now has an easyCore compatible form

            :param x: array of data points to be calculated
            :type x: np.ndarray
            :param kwargs: key word arguments
            :return: points calculated at `x`
            :rtype: np.ndarray
            """
            # Update the `Parameter` values and the callback if needed
            # TODO THIS IS NOT THREAD SAFE :-(
            for name, value in kwargs.items():
                par_name = int(name[1:])
                if par_name in self._cached_pars.keys():
                    # This will take in to account constraints
                    self._cached_pars[par_name].value = value
                    # Since we are calling the parameter fset will be called.
            # TODO Pre processing here
            for constraint in self.fit_constraints():
                constraint()
            return_data = func(x)
            # TODO Loading or manipulating data here
            return return_data

        self._fit_function = fit_function
        return fit_function

    def fit(self, x: np.ndarray, y: np.ndarray, weights: Optional[np.ndarray] = None,
            model=None, parameters=None, method: str = None, xtol: float = 1e-6, ftol: float = 1e-8,
            **kwargs) -> FitResults:
        """
        Perform a fit using the DFO-ls engine.

        :param x: points to be calculated at
        :type x: np.ndarray
        :param y: measured points
        :type y: np.ndarray
        :param weights: Weights for supplied measured points * Not really optional*
        :type weights: np.ndarray
        :param model: Optional Model which is being fitted to
        :type model: lmModel
        :param parameters: Optional parameters for the fit
        :type parameters: List[bumpsParameter]
        :param kwargs: Additional arguments for the fitting function.
        :param method: Method for minimization
        :type method: str
        :return: Fit results
        :rtype: ModelResult
        """

        default_method = {}
        if method is not None and method in self.available_methods():
            default_method['method'] = method

        if weights is None:
            weights = np.sqrt(np.abs(y))

        if model is None:
            model = self.make_model(pars=parameters)
            model = model(x, y, weights)
        self._cached_model = model
        self.p_0 = {f'p{key}': self._cached_pars[key].raw_value for key in self._cached_pars.keys()}

        # Why do we do this? Because a fitting template has to have borg instantiated outside pre-runtime
        from easyCore import borg
        borg.stack.beginMacro('Fitting routine')
        try:
            model_results = self.dfols_fit(model, **kwargs)
            self._set_parameter_fit_result(model_results)
            results = self._gen_fit_results(model_results)
        except Exception as e:
            raise FitError(e)
        finally:
            borg.stack.endMacro()
        return results

    def convert_to_pars_obj(self, par_list: Optional[list] = None):
        """
        NOTE THAT THIS IS NOT NEEDED FOR DFO-LS
        """

        pass

    @staticmethod
    def convert_to_par_object(obj) -> None:
        """
        Convert an `easyCore.Objects.Base.Parameter` object to a new Parameter object
        NOTE THAT THIS IS NOT NEEDED FOR DFO-LS
        """
        pass

    def _set_parameter_fit_result(self, fit_result, ci: float = 0.95) -> None:
        """
        Update parameters to their final values and assign a std error to them.

        :param fit_result: Fit object which contains info on the fit
        :param ci: Confidence interval for calculating errors. Default 95%
        :return: None
        :rtype: noneType
        """
        pars = self._cached_pars
        error_matrix = self._error_from_jacobian(fit_result.jacobian, fit_result.resid, ci)
        for idx, par in enumerate(pars.values()):
            par.value = fit_result.x[idx]
            par.error = error_matrix[idx, idx]

    def _gen_fit_results(self, fit_results, **kwargs) -> FitResults:
        """
        Convert fit results into the unified `FitResults` format

        :param fit_result: Fit object which contains info on the fit
        :return: fit results container
        :rtype: FitResults
        """

        results = FitResults()
        for name, value in kwargs.items():
            if getattr(results, name, False):
                setattr(results, name, value)
        results.success = not bool(fit_results.flag)
        pars = self._cached_pars
        item = {}
        for p_name, par in pars.items():
            item[f'p{p_name}'] = par.raw_value
        results.p0 = self.p_0
        results.p = item
        results.x = self._cached_model.x
        results.y_obs = self._cached_model.y
        results.y_calc = self.evaluate(results.x, parameters=results.p)
        results.residual = results.y_obs - results.y_calc
        results.goodness_of_fit = fit_results.f

        results.fitting_engine = self.__class__
        results.fit_args = None
        # results.check_sanity()

        return results

    def available_methods(self) -> List[str]:
        return ['leastsq']

    def dfols_fit(self, model: Callable, **kwargs):
        """
        Method to convert easyCore styling to DFO-LS styling (yes, again)

        :param model: Model which accepts f(x[0])
        :type model: Callable
        :param kwargs: Any additional arguments for dfols.solver
        :type kwargs: dict
        :return: dfols fit results container
=        """
        x0 = np.array([par.raw_value for par in iter(self._cached_pars.values())])
        bounds = (np.array([par.min for par in iter(self._cached_pars.values())]),
                  np.array([par.max for par in iter(self._cached_pars.values())]))
        results = dfols.solve(model, x0, bounds=bounds, **kwargs)
        return results



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

import sys

from typing import Callable, List, Any
from functools import wraps

from easyCore import borg
from easyCore.Utils.Hugger.Hugger import Store, PatcherFactory


class LoggedProperty(property):
    """
    Pump up python properties. In this case we can see who has called this property and
    then do something if a criteria is met. In this case if the caller is not a member of
    the `BaseObj` class. Note that all high level `easyCore` objects should be built from
    `BaseObj`.
    """

    _borg = borg

    def __init__(self, *args, get_id=None, my_self=None, test_class=None, **kwargs):
        super(LoggedProperty, self).__init__(*args, **kwargs)
        self._get_id = get_id
        self._my_self = my_self
        self.test_class = test_class

    @staticmethod
    def _caller_class(test_class, skip: int = 1):
        def stack_(frame):
            frame_list = []
            while frame:
                frame_list.append(frame)
                frame = frame.f_back
            return frame_list

        stack: List[Any] = stack_(sys._getframe(1))
        start = 0 + skip
        if len(stack) < start + 1:
            return ""
        parent_frame = stack[start]
        test = False
        if "self" in parent_frame.f_locals:
            test = issubclass(parent_frame.f_locals["self"].__class__, test_class)
        return test

    def __get__(self, instance, owner=None):
        if not borg.script.enabled:
            return super(LoggedProperty, self).__get__(instance, owner)
        test = self._caller_class(self.test_class)
        res = super(LoggedProperty, self).__get__(instance, owner)

        def result_item(item_to_be_resulted):
            if item_to_be_resulted is None:
                return None
            if borg.map.is_known(item_to_be_resulted):
                borg.map.change_type(item_to_be_resulted, "returned")
            else:
                borg.map.add_vertex(item_to_be_resulted, obj_type="returned")

        if not test and self._get_id is not None and self._my_self is not None:
            if not isinstance(res, list):
                result_item(res)
            else:
                for item in res:
                    result_item(item)
            Store().append_log(self.makeEntry("get", res))
            if borg.debug:  # noqa: S1006
                print(
                    f"I'm {self._my_self} and {self._get_id} has been called from the outside!"
                )
        return res

    def __set__(self, instance, value):
        if not borg.script.enabled:
            return super().__set__(instance, value)
        test = self._caller_class(self.test_class)
        if not test and self._get_id is not None and self._my_self is not None:
            Store().append_log(self.makeEntry("set", value))
            if borg.debug:  # noqa: S1006
                print(
                    f"I'm {self._my_self} and {self._get_id} has been set to {value} from the outside!"
                )
        return super().__set__(instance, value)

    def makeEntry(self, log_type, returns, *args, **kwargs) -> str:
        temp = ""
        if returns is None:
            returns = []
        if not isinstance(returns, list):
            returns = [returns]
        if log_type == "get":
            for var in returns:
                if borg.map.convert_id_to_key(var) in borg.map.returned_objs:
                    index = borg.map.returned_objs.index(
                        borg.map.convert_id_to_key(var)
                    )
                    temp += f"{Store().var_ident}{index}, "
            if len(returns) > 0:
                temp = temp[:-2]
                temp += " = "
            if borg.map.convert_id_to_key(self._my_self) in borg.map.created_objs:
                # for edge in route[::-1]:
                index = borg.map.created_objs.index(
                    borg.map.convert_id_to_key(self._my_self)
                )
                temp += (
                    f"{self._my_self.__class__.__name__.lower()}_{index}.{self._get_id}"
                )
            if borg.map.convert_id(self._my_self) in borg.map.created_internal:
                # We now have to trace....
                route = borg.map.reverse_route(self._my_self)  # noqa: F841
                index = borg.map.created_objs.index(
                    borg.map.convert_id_to_key(self._my_self)
                )
                temp += (
                    f"{self._my_self.__class__.__name__.lower()}_{index}.{self._get_id}"
                )
        elif log_type == "set":
            if borg.map.convert_id_to_key(self._my_self) in borg.map.created_objs:
                index = borg.map.created_objs.index(
                    borg.map.convert_id_to_key(self._my_self)
                )
                temp += f"{self._my_self.__class__.__name__.lower()}_{index}.{self._get_id} = "
            args = args[1:]
            for var in args:
                if borg.map.convert_id_to_key(var) in borg.map.argument_objs:
                    index = borg.map.argument_objs.index(
                        borg.map.convert_id_to_key(var)
                    )
                    temp += f"{Store().var_ident}{index}"
                elif borg.map.convert_id_to_key(var) in borg.map.returned_objs:
                    index = borg.map.returned_objs.index(
                        borg.map.convert_id_to_key(var)
                    )
                    temp += f"{Store().var_ident}{index}"
                elif borg.map.convert_id_to_key(var) in borg.map.created_objs:
                    index = borg.map.created_objs.index(borg.map.convert_id_to_key(var))
                    temp += f"{self._my_self.__class__.__name__.lower()}_{index}"
                else:
                    if isinstance(var, str):
                        var = '"' + var + '"'
                    temp += f"{var}"
        else:
            print(f"{log_type} is not implemented yet. Sorry")
        temp += "\n"
        return temp


class PropertyHugger(PatcherFactory):
    # Properties are immutable, so need to be set at the parent level. However unlike `FunctionHugger` we can't traverse
    # the stack to get the parent. So, it and it's name has to be set at initialization. Boo!

    _borg = borg

    def __init__(self, klass, prop_name):
        super().__init__()
        self.klass = klass
        if isinstance(prop_name, tuple):
            self.prop_name = prop_name[0]
            self.property = prop_name[1]
        else:
            self.prop_name = prop_name
            self.property = klass.__dict__.get(prop_name)
        self.__patch_ref = {
            "fget": {"old": self.property.fget, "patcher": self.patch_get},
            "fset": {"old": self.property.fset, "patcher": self.patch_set},
            "fdel": {"old": self.property.fdel, "patcher": self.patch_del},
        }

    def patch(self):
        option = {}
        for key, item in self.__patch_ref.items():
            func = getattr(self.property, key)
            if func is not None:
                if borg.debug:
                    print(f"Patching property {self.klass.__name__}.{self.prop_name}")
                patch_function: Callable = item.get("patcher")
                new_func = patch_function(func)
                option[key] = new_func
        setattr(self.klass, self.prop_name, property(**option))

    def restore(self):
        if borg.debug:
            print(f"Restoring property {self.klass.__name__}.{self.prop_name}")
        setattr(self.klass, self.prop_name, self.property)

    def patch_get(self, func: Callable) -> Callable:
        @wraps(func)
        def inner(*args, **kwargs):
            if borg.debug:
                print(
                    f"{self.klass.__name__}.{self.prop_name} has been called with {args[1:]}, {kwargs}"
                )
            res = func(*args, **kwargs)
            self._append_args(*args, **kwargs)
            self._append_result(res)
            self._append_log(self.makeEntry("get", res, *args, **kwargs))
            return res

        return inner

    def patch_set(self, func: Callable) -> Callable:
        @wraps(func)
        def inner(*args, **kwargs):
            if borg.debug:
                print(
                    f"{self.klass.__name__}.{self.prop_name} has been set with {args[1:]}, {kwargs}"
                )
            self._append_args(*args, **kwargs)
            self._append_log(self.makeEntry("set", None, *args, **kwargs))
            return func(*args, **kwargs)

        return inner

    def patch_del(self, func: Callable) -> Callable:
        @wraps(func)
        def inner(*args, **kwargs):
            if borg.debug:
                print(f"{self.klass.__name__}.{self.prop_name} has been deleted.")
            self._append_log(self.makeEntry("del", None, *args, **kwargs))
            return func(*args, **kwargs)

        return inner

    def makeEntry(self, log_type, returns, *args, **kwargs) -> str:
        temp = ""
        if returns is None:
            returns = []
        if not isinstance(returns, list):
            returns = [returns]
        if log_type == "get":
            for var in returns:
                if self._in_list("return_list", var):
                    index = self._get_position("return_list", var)
                    temp += f"{self._store.var_ident}{index}, "
            if len(returns) > 0:
                temp = temp[:-2]
                temp += " = "
            if self._in_list("create_list", args[0]):
                index = self._get_position("create_list", args[0])
                temp += f"{self.klass.__name__.lower()}_{index}.{self.prop_name}"
        elif log_type == "set":
            if self._in_list("create_list", args[0]):
                index = self._get_position("create_list", args[0])
                temp += f"{self.klass.__name__.lower()}_{index}.{self.prop_name} = "
            args = args[1:]
            for var in args:
                if self._in_list("input_list", var):
                    index = self._get_position("input_list", var)
                    temp += f"{self._store.var_ident}{index}"
                elif self._in_list("return_list", var):
                    index = self._get_position("return_list", var)
                    temp += f"{self._store.var_ident}{index}"
                elif self._in_list("create_list", var):
                    index = self._get_position("create_list", var)
                    temp += f"{self.klass.__name__.lower()}_{index}"
                else:
                    if isinstance(var, str):
                        var = '"' + var + '"'
                    temp += f"{var}"
        else:
            print(f"{log_type} is not implemented yet. Sorry")
        temp += "\n"
        return temp



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

import logging


class Logger:
    def __init__(self, log_level: int = logging.INFO):
        self.logger = logging.getLogger(__name__)
        self.level = log_level
        self.logger.setLevel(self.level)

    def getLogger(self, logger_name, color: str = '32', defaults: bool = True) -> logging:
        """
        Create a logger
        :param color:
        :param logger_name: logger name. Usually __name__ on creation
        :param defaults: Do you want to associate any current file loggers with this logger
        :return: A logger
        """
        logger = logging.getLogger(logger_name)
        logger.setLevel(self.level)
        # self.applyLevel(logger)
        # for handler_type in self._handlers:
        #     for handler in self._handlers[handler_type]:
        #         if handler_type == 'sys' or defaults:
        #             handler.formatter._fmt = self._makeColorText(color)
        #             logger.addHandler(handler)
        # logger.propagate = False
        # self._loggers.append(logger)
        return logger



__author__ = "https://github.com/materialsvirtuallab/monty/blob/master/monty/json.py"
__version__ = "3.0.0"

from typing import Optional, List, Dict, Any

"""
This code has been taken and modified from https://guide.materialsvirtuallab.org/monty/index.html
Many thanks to the original authors, the source and contributors can be found at __author__

The MIT License (MIT)

Copyright (c) 2014 Materials Virtual Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


JSON serialization and deserialization utilities.
"""

#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

import os
import json
import datetime

from hashlib import sha1
from collections import OrderedDict, defaultdict
from enum import Enum

from importlib import import_module

from inspect import getfullargspec

from easyCore import np


_KNOWN_CORE_TYPES = ("Descriptor", "Parameter")

try:
    import bson
except ImportError:
    bson = None

try:
    import ruamel.yaml as yaml
except ImportError:
    try:
        import yaml  # type: ignore
    except ImportError:
        yaml = None  # type: ignore


def _load_redirect(redirect_file):
    try:
        with open(redirect_file, "rt") as f:
            d = yaml.safe_load(f)
    except IOError:
        # If we can't find the file
        # Just use an empty redirect dict
        return {}

    # Convert the full paths to module/class
    redirect_dict = defaultdict(dict)
    for old_path, new_path in d.items():
        old_class = old_path.split(".")[-1]
        old_module = ".".join(old_path.split(".")[:-1])

        new_class = new_path.split(".")[-1]
        new_module = ".".join(new_path.split(".")[:-1])

        redirect_dict[old_module][old_class] = {
            "@module": new_module,
            "@class": new_class,
        }

    return dict(redirect_dict)


def get_class_module(obj):
    """
    Returns the REAL module of the class of the object.
    """
    c = getattr(obj, "__old_class__", obj.__class__)
    return c.__module__


class MSONable:
    """
    This is a mix-in base class specifying an API for msonable objects. MSON
    is Monty JSON. Essentially, MSONable objects must implement an as_dict
    method, which must return a json serializable dict and must also support
    no arguments (though optional arguments to finetune the output is ok),
    and a from_dict class method that regenerates the object from the dict
    generated by the as_dict method. The as_dict method should contain the
    "@module" and "@class" keys which will allow the MontyEncoder to
    dynamically deserialize the class. E.g.::

        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__

    A default implementation is provided in MSONable, which automatically
    determines if the class already contains self.argname or self._argname
    attributes for every arg. If so, these will be used for serialization in
    the dict format. Similarly, the default from_dict will deserialization
    classes of such form. An example is given below::

        class MSONClass(MSONable):

        def __init__(self, a, b, c, d=1, **kwargs):
            self.a = a
            self.b = b
            self._c = c
            self._d = d
            self.kwargs = kwargs

    For such classes, you merely need to inherit from MSONable and you do not
    need to implement your own as_dict or from_dict protocol.

    New to Monty V2.0.6....
    Classes can be redirected to moved implementations by putting in the old
    fully qualified path and new fully qualified path into .monty.yaml in the
    home folder

    Example:
    old_module.old_class: new_module.new_class
    """

    REDIRECT = _load_redirect(os.path.join(os.path.expanduser("~"), ".monty.yaml"))

    def as_dict(self, skip: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        A JSON serializable dict representation of an object.
        """
        if skip is None:
            skip = []
        elif isinstance(skip, str):
            skip = [skip]
        if not isinstance(skip, list):
            raise ValueError("Skip must be a list of strings.")

        d = {"@module": get_class_module(self), "@class": self.__class__.__name__}

        try:
            parent_module = get_class_module(self).split(".")[0]
            module_version = import_module(parent_module).__version__  # type: ignore
            d["@version"] = "{}".format(module_version)
        except (AttributeError, ImportError):
            d["@version"] = None  # type: ignore

        spec = getfullargspec(self.__class__.__init__)
        args = spec.args

        def recursive_as_dict(obj):
            if isinstance(obj, (list, tuple)):
                return [recursive_as_dict(it) for it in obj]
            if isinstance(obj, dict):
                return {kk: recursive_as_dict(vv) for kk, vv in obj.items()}
            if hasattr(obj, "as_dict"):
                return obj.as_dict(skip=skip)
            return obj

        for c in args:
            if c != "self" and c not in skip:
                try:
                    a = self.__getattribute__(c)
                except AttributeError:
                    try:
                        a = self.__getattribute__("_" + c)
                    except AttributeError:
                        err = True
                        if hasattr(self, "kwargs"):
                            # type: ignore
                            option = getattr(self, "kwargs")
                            if hasattr(option, c):
                                v = getattr(option, c)
                                delattr(option, c)
                                d.update(v)  # pylint: disable=E1101
                                err = False
                        if hasattr(self, "_kwargs"):
                            # type: ignore
                            option = getattr(self, "_kwargs")
                            if hasattr(option, c):
                                v = getattr(option, c)
                                delattr(option, c)
                                d.update(v)  # pylint: disable=E1101
                                err = False
                        if err:
                            raise NotImplementedError(
                                "Unable to automatically determine as_dict "
                                "format from class. MSONAble requires all "
                                "args to be present as either self.argname or "
                                "self._argname, and kwargs to be present under"
                                "a self.kwargs variable to automatically "
                                "determine the dict format. Alternatively, "
                                "you can implement both as_dict and from_dict."
                            )
                d[c] = recursive_as_dict(a)
        if hasattr(self, "kwargs"):
            # type: ignore
            d.update(
                {k: v for k, v in getattr(self, "kwargs").items() if k not in skip}
            )  # pylint: disable=E1101
        if spec.varargs is not None and getattr(self, spec.varargs, None) is not None:
            d.update({spec.varargs: getattr(self, spec.varargs)})
        if hasattr(self, "_kwargs"):
            d.update(
                {k: v for k, v in getattr(self, "_kwargs").items() if k not in skip}
            )  # pylint: disable=E1101
        if isinstance(self, Enum):
            d.update({"value": self.value})  # pylint: disable=E1101
        return d

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "MSONable":
        """
        :param d: Dict representation.
        :return: MSONable class.
        """
        decoded = {
            k: MontyDecoder().process_decoded(v)
            for k, v in d.items()
            if not k.startswith("@")
        }
        return cls(**decoded)

    def to_json(self) -> str:
        """
        Returns a json string representation of the MSONable object.
        """
        return json.dumps(self, cls=MontyEncoder)

    @staticmethod
    def __parse_dict(in_dict: Dict[str, Any]) -> Dict[str, Any]:
        out_dict = dict()
        for key in in_dict.keys():
            if key[0] == "@":
                if key == "@class" and in_dict[key] not in _KNOWN_CORE_TYPES:
                    out_dict["name"] = in_dict[key]
                continue
            out_dict[key] = in_dict[key]
            if isinstance(in_dict[key], dict):
                out_dict[key] = MSONable.__parse_dict(in_dict[key])
            elif isinstance(in_dict[key], list):
                out_dict[key] = [
                    MSONable.__parse_dict(x) if isinstance(x, dict) else x
                    for x in in_dict[key]
                ]
        return out_dict

    def to_data_dict(self, skip: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Convert the object to a reduced dictionary only containing data.

        :param skip: Which items should be skipped.
        :return: Dictionary representation of the object.
        """
        if skip is None:
            skip = []
        elif isinstance(skip, str):
            skip = [skip]
        if not isinstance(skip, list):
            raise ValueError("Skip must be a list of strings.")

        d = self.as_dict(skip=skip)
        return self.__parse_dict(d)

    def unsafe_hash(self) -> sha1:
        """
        Returns an hash of the current object. This uses a generic but low
        performance method of converting the object to a dictionary, flattening
        any nested keys, and then performing a hash on the resulting object
        """

        def flatten(obj, seperator="."):
            # Flattens a dictionary

            flat_dict = {}
            for key, value in obj.items():
                if isinstance(value, dict):
                    flat_dict.update(
                        {
                            seperator.join([key, _key]): _value
                            for _key, _value in flatten(value).items()
                        }
                    )
                elif isinstance(value, list):
                    list_dict = {
                        "{}{}{}".format(key, seperator, num): item
                        for num, item in enumerate(value)
                    }
                    flat_dict.update(flatten(list_dict))
                else:
                    flat_dict[key] = value

            return flat_dict

        ordered_keys = sorted(
            flatten(jsanitize(self.as_dict())).items(), key=lambda x: x[0]
        )
        ordered_keys = [item for item in ordered_keys if "@" not in item[0]]
        return sha1(json.dumps(OrderedDict(ordered_keys)).encode("utf-8"))  # nosec


class MontyEncoder(json.JSONEncoder):
    """
    A Json Encoder which supports the MSONable API, plus adds support for
    numpy arrays, datetime objects, bson ObjectIds (requires bson).

    Usage::

        # Add it as a *cls* keyword when using json.dump
        json.dumps(object, cls=MontyEncoder)
    """

    def default(self, o) -> dict:  # pylint: disable=E0202
        """
        Overriding default method for JSON encoding. This method does two
        things: (a) If an object has a to_dict property, return the to_dict
        output. (b) If the @module and @class keys are not in the to_dict,
        add them to the output automatically. If the object has no to_dict
        property, the default Python json encoder default method is called.

        Args:
            o: Python object.

        Return:
            Python dict representation.
        """
        if isinstance(o, datetime.datetime):
            return {"@module": "datetime", "@class": "datetime", "string": o.__str__()}
        if np is not None:
            if isinstance(o, np.ndarray):
                if str(o.dtype).startswith("complex"):
                    return {
                        "@module": "numpy",
                        "@class": "array",
                        "dtype": o.dtype.__str__(),
                        "data": [o.real.tolist(), o.imag.tolist()],
                    }
                return {
                    "@module": "numpy",
                    "@class": "array",
                    "dtype": o.dtype.__str__(),
                    "data": o.tolist(),
                }
            if isinstance(o, np.generic):
                return o.item()
        if bson is not None:
            if isinstance(o, bson.objectid.ObjectId):
                return {"@module": "bson.objectid", "@class": "ObjectId", "oid": str(o)}

        try:
            d = o.as_dict()
            if "@module" not in d:
                d["@module"] = "{}".format(o.__class__.__module__)
            if "@class" not in d:
                d["@class"] = "{}".format(o.__class__.__name__)
            if "@version" not in d:
                try:
                    parent_module = o.__class__.__module__.split(".")[0]
                    module_version = import_module(parent_module).__version__  # type: ignore
                    d["@version"] = "{}".format(module_version)
                except (AttributeError, ImportError):
                    d["@version"] = None
            return d
        except AttributeError:
            return json.JSONEncoder.default(self, o)


class MontyDecoder(json.JSONDecoder):
    """
    A Json Decoder which supports the MSONable API. By default, the
    decoder attempts to find a module and name associated with a dict. If
    found, the decoder will generate a Pymatgen as a priority.  If that fails,
    the original decoded dictionary from the string is returned. Note that
    nested lists and dicts containing pymatgen object will be decoded correctly
    as well.

    Usage:

        # Add it as a *cls* keyword when using json.load
        json.loads(json_string, cls=MontyDecoder)
    """

    def process_decoded(self, d):
        """
        Recursive method to support decoding dicts and lists containing
        pymatgen objects.
        """
        if isinstance(d, dict):
            if "@module" in d and "@class" in d:
                modname = d["@module"]
                classname = d["@class"]
                if classname in MSONable.REDIRECT.get(modname, {}):
                    modname = MSONable.REDIRECT[modname][classname]["@module"]
                    classname = MSONable.REDIRECT[modname][classname]["@class"]
            else:
                modname = None
                classname = None
            if modname and modname not in ["bson.objectid", "numpy"]:
                if modname == "datetime" and classname == "datetime":
                    try:
                        dt = datetime.datetime.strptime(
                            d["string"], "%Y-%m-%d %H:%M:%S.%f"
                        )
                    except ValueError:
                        dt = datetime.datetime.strptime(
                            d["string"], "%Y-%m-%d %H:%M:%S"
                        )
                    return dt

                mod = __import__(modname, globals(), locals(), [classname], 0)
                if hasattr(mod, classname):
                    cls_ = getattr(mod, classname)
                    data = {k: v for k, v in d.items() if not k.startswith("@")}
                    if hasattr(cls_, "from_dict"):
                        return cls_.from_dict(data)
            elif np is not None and modname == "numpy" and classname == "array":
                if d["dtype"].startswith("complex"):
                    return np.array(
                        [r + i * 1j for r, i in zip(*d["data"])], dtype=d["dtype"]
                    )
                return np.array(d["data"], dtype=d["dtype"])

            elif (
                (bson is not None)
                and modname == "bson.objectid"
                and classname == "ObjectId"
            ):
                return bson.objectid.ObjectId(d["oid"])

            return {
                self.process_decoded(k): self.process_decoded(v) for k, v in d.items()
            }

        if isinstance(d, list):
            return [self.process_decoded(x) for x in d]

        return d

    def decode(self, s):
        """
        Overrides decode from JSONDecoder.

        :param s: string
        :return: Object.
        """
        d = json.JSONDecoder.decode(self, s)
        return self.process_decoded(d)


class MSONError(Exception):
    """
    Exception class for serialization errors.
    """


def jsanitize(obj, strict=False, allow_bson=False):
    """
    This method cleans an input json-like object, either a list or a dict or
    some sequence, nested or otherwise, by converting all non-string
    dictionary keys (such as int and float) to strings, and also recursively
    encodes all objects using Monty's as_dict() protocol.

    Args:
        obj: input json-like object.
        strict (bool): This parameters sets the behavior when jsanitize
            encounters an object it does not understand. If strict is True,
            jsanitize will try to get the as_dict() attribute of the object. If
            no such attribute is found, an attribute error will be thrown. If
            strict is False, jsanitize will simply call str(object) to convert
            the object to a string representation.
        allow_bson (bool): This parameters sets the behavior when jsanitize
            encounters an bson supported type such as objectid and datetime. If
            True, such bson types will be ignored, allowing for proper
            insertion into MongoDb databases.

    Returns:
        Sanitized dict that can be json serialized.
    """
    if allow_bson and (
        isinstance(obj, (datetime.datetime, bytes))
        or (bson is not None and isinstance(obj, bson.objectid.ObjectId))
    ):
        return obj
    if isinstance(obj, (list, tuple)):
        return [jsanitize(i, strict=strict, allow_bson=allow_bson) for i in obj]
    if np is not None and isinstance(obj, np.ndarray):
        return [
            jsanitize(i, strict=strict, allow_bson=allow_bson) for i in obj.tolist()
        ]
    if isinstance(obj, dict):
        return {
            k.__str__(): jsanitize(v, strict=strict, allow_bson=allow_bson)
            for k, v in obj.items()
        }
    if isinstance(obj, (int, float)):
        return obj
    if obj is None:
        return None

    if not strict:
        return obj.__str__()

    if isinstance(obj, str):
        return obj.__str__()

    return jsanitize(obj.as_dict(), strict=strict, allow_bson=allow_bson)



__author__ = "github.com/wardsimon"
__version__ = "0.0.1"

#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from abc import ABCMeta
from types import FunctionType
from typing import List, Callable, TypeVar


from easyCore import borg, default_fitting_engine
import easyCore.Fitting as Fitting

_C = TypeVar("_C", bound=ABCMeta)
_M = TypeVar("_M", bound=Fitting.FittingTemplate)


class Fitter:
    """
    Wrapper to the fitting engines
    """

    _borg = borg

    def __init__(self, fit_object: object = None, fit_function: Callable = None):

        self._fit_object = fit_object
        self._fit_function = fit_function

        can_initialize = False
        # We can only proceed if both obj and func are not None
        if (fit_object is not None) & (fit_function is not None):
            can_initialize = True
        else:
            if (fit_object is not None) or (fit_function is not None):
                raise AttributeError

        self._engines: List[_C] = Fitting.engines
        self._current_engine: _C = None
        self.__engine_obj: _M = None
        self._is_initialized: bool = False
        self.create()

        fit_methods = [
            x
            for x, y in Fitting.FittingTemplate.__dict__.items()
            if isinstance(y, FunctionType) and not x.startswith("_")
        ]
        for method_name in fit_methods:
            setattr(self, method_name, self.__pass_through_generator(method_name))

        if can_initialize:
            self.__initialize()

    def initialize(self, fit_object: object, fit_function: Callable):
        self._fit_object = fit_object
        self._fit_function = fit_function
        self.__initialize()

    def __initialize(self):
        self.__engine_obj = self._current_engine(self._fit_object, self._fit_function)
        self._is_initialized = True

    def create(self, engine_name: str = default_fitting_engine):
        engines = self.available_engines
        if engine_name in engines:
            self._current_engine = self._engines[engines.index(engine_name)]

    def switch_engine(self, engine_name: str):
        # There isn't any state to carry over
        if not self._is_initialized:
            print("The fitting engine must first be initialized")
            raise ReferenceError
        constraints = self.__engine_obj._constraints
        self.create(engine_name)
        self.__initialize()
        self.__engine_obj._constraints = constraints

    @property
    def available_engines(self) -> List[str]:
        """
        Get a list of the names of available fitting engines

        :return: List of available fitting engines
        :rtype: List[str]
        """
        if Fitting.engines is None:
            print("Fitting not instantiated yet")
            raise ImportError
        return [engine.name for engine in Fitting.engines]

    @property
    def can_fit(self) -> bool:
        """
        Can a fit be performed. i.e has the object been created properly

        :return: Can a fit be performed
        :rtype: bool
        """
        return self._is_initialized

    @property
    def current_engine(self) -> _C:
        """
        Get the class object of the current fitting engine.

        :return: Class of the current fitting engine (based on the `FittingTemplate` class)
        :rtype: _T
        """
        return self._current_engine

    @property
    def engine(self) -> _M:
        """
        Get the current fitting engine object.

        :return:
        :rtype: _M
        """
        return self.__engine_obj

    @property
    def fit_function(self) -> Callable:
        return self._fit_function

    @fit_function.setter
    def fit_function(self, fit_function: Callable):
        self._fit_function = fit_function
        self.__initialize()

    @property
    def fit_object(self) -> object:
        return self._fit_object

    @fit_object.setter
    def fit_object(self, fit_object: object):
        self._fit_object = fit_object
        self.__initialize()

    def __pass_through_generator(self, name: str):
        obj = self

        def inner(*args, **kwargs):
            if not obj.can_fit:
                raise ReferenceError("The fitting engine must first be initialized")
            func = getattr(obj.engine, name, None)
            if func is None:
                raise ValueError
            return func(*args, **kwargs)

        return inner



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easyCore',
 'easyCore.Datasets',
 'easyCore.Fitting',
 'easyCore.Objects',
 'easyCore.Utils',
 'easyCore.Utils.Hugger',
 'easyCore.Utils.io',
 'easyCore.models']

package_data = \
{'': ['*']}

install_requires = \
['asteval>=0.9.27,<0.10.0',
 'bumps>=0.8,<0.10',
 'lmfit>=1.0,<2.0',
 'numpy>=1.21,<2.0',
 'pint>=0.17,<0.19',
 'uncertainties>=3.1,<4.0',
 'xarray>=0.16,<0.21']

setup_kwargs = {
    'name': 'easysciencecore',
    'version': '0.2.3',
    'description': 'Generic logic for easyScience libraries',
    'long_description': '# [![License][50]][51] [![Release][32]][33] [![Downloads][70]][71] [![CI Build][20]][21] \n\n[![CodeFactor][83]][84] [![Codecov][85]][86]\n[![Lines of code][81]](<>) [![Total lines][80]](<>) [![Files][82]](<>)\n\n\n<img height="80"><img src="https://raw.githubusercontent.com/easyScience/easyCore/master/resources/images/ec_logo.svg" height="65">\n\n**easyCore** is the foundation of the *easyScience* universe, providing the building blocks for libraries and applications which aim to make scientific data simulation and analysis easier.\n\n## Install\n\n**easyCore** can be downloaded using pip:\n\n```pip install easysciencecore```\n\nOr direct from the repository:\n\n```pip install https://github.com/easyScience/easyCore```\n\n## Test\n\nAfter installation, launch the test suite:\n\n```python -m pytest```\n\n## Documentation\n\nDocumentation can be found at:\n\n[https://easyScience.github.io/easyCore](https://easyScience.github.io/easyCore)\n\n## Contributing\nWe absolutely welcome contributions. **easyCore** is maintained by the ESS and on a volunteer basis and thus we need to foster a community that can support user questions and develop new features to make this software a useful tool for all users while encouraging every member of the community to share their ideas.\n\n## License\nWhile **easyCore** is under the BSD-3 license, DFO_LS is subject to the GPL license.\n\n<!---CI Build Status--->\n\n[20]: https://github.com/easyScience/easyCore/workflows/CI%20using%20pip/badge.svg\n\n[21]: https://github.com/easyScience/easyCore/actions\n\n\n<!---Release--->\n\n[32]: https://img.shields.io/pypi/v/easysciencecore?color=green\n\n[33]: https://pypi.org/project/easyScienceCore\n\n\n<!---License--->\n\n[50]: https://img.shields.io/github/license/easyScience/easyCore.svg\n\n[51]: https://github.com/easyScience/easyCore/blob/master/LICENSE.md\n\n\n<!---Downloads--->\n\n[70]: https://img.shields.io/pypi/dm/easyScienceCore.svg\n\n[71]: https://pypi.org/project/easyScienceCore\n\n<!---Code statistics--->\n\n[80]: https://tokei.rs/b1/github/easyScience/easyCore\n\n[81]: https://tokei.rs/b1/github/easyScience/easyCore?category=code\n\n[82]: https://tokei.rs/b1/github/easyScience/easyCore?category=files\n\n[83]: https://www.codefactor.io/repository/github/easyscience/easycore/badge\n\n[84]: https://www.codefactor.io/repository/github/easyscience/easycore\n\n[85]: https://img.shields.io/codecov/c/github/easyScience/easyCore?color=green\n\n[86]: https://app.codecov.io/gh/easyScience/easyCore/',
    'author': 'Simon Ward',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/easyScience/easyCore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'


#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

from abc import ABCMeta, abstractmethod
from typing import Union, Callable, List, Optional

from easyCore import np
from scipy import stats


class FittingTemplate(metaclass=ABCMeta):
    """
    This template class is the basis for all fitting engines in `easyCore`.
    """

    _engines = []
    property_type = None
    name: str = ''

    def __init_subclass__(cls, is_abstract: bool = False, **kwargs):
        super().__init_subclass__(**kwargs)
        if not is_abstract:
            # Deal with the issue of people not reading the schema.
            if not hasattr(cls, 'name'):
                setattr(cls, 'name', cls.__class__.__name__)
            cls._engines.append(cls)

    def __init__(self, obj, fit_function: Callable):
        self._object = obj
        self._original_fit_function = fit_function
        self._cached_pars = {}
        self._cached_model = None
        self._fit_function = None
        self._constraints = []
        self._dataset = None

    @property
    def all_constraints(self) -> list:
        return [*self._constraints, *self._object._constraints]

    def fit_constraints(self) -> list:
        return self._constraints

    def add_fit_constraint(self, constraint):
        self._constraints.append(constraint)

    def remove_fit_constraint(self, index: int):
        del self._constraints[index]

    @abstractmethod
    def make_model(self, pars=None):
        """
        Generate an engine model from the supplied `fit_function` and parameters in the base object.

        :return: Callable model
        """

    @abstractmethod
    def _generate_fit_function(self) -> Callable:
        """
        Using the user supplied `fit_function`, wrap it in such a way we can update `Parameter` on
        iterations.
        """

    @abstractmethod
    def fit(self, x: np.ndarray, y: np.ndarray,
            weights: Optional[Union[np.ndarray]] = None,
            model: Optional = None, parameters: Optional = None,
            method: Optional = None, **kwargs):
        """
        Perform a fit using the  engine.

        :param x: points to be calculated at
        :type x: np.ndarray
        :param y: measured points
        :type y: np.ndarray
        :param weights: Weights for supplied measured points
        :type weights: np.ndarray
        :param model: Optional Model which is being fitted to
        :param parameters: Optional parameters for the fit
        :param method: method for the minimizer to use.
        :type method: str
        :param kwargs: Additional arguments for the fitting function.
        :return: Fit results
        """

    def evaluate(self, x: np.ndarray, parameters: dict = None, **kwargs) -> np.ndarray:
        """
        Evaluate the fit function for values of x. Parameters used are either the latest or user supplied.
        If the parameters are user supplied, it must be in a dictionary of {'parameter_name': parameter_value,...}.

        :param x: x values for which the fit function will be evaluated
        :type x:  np.ndarray
        :param parameters: Dictionary of parameters which will be used in the fit function. They must be in a dictionary
         of {'parameter_name': parameter_value,...}
        :type parameters: dict
        :param kwargs: additional arguments
        :return: y values calculated at points x for a set of parameters.
        :rtype: np.ndarray
        """
        if self._fit_function is None:
            # This will also generate self._cached_pars
            self._fit_function = self._generate_fit_function()

        if not isinstance(parameters, (dict, type(None))):
            raise AttributeError

        pars = self._cached_pars
        new_parameters = parameters
        if new_parameters is None:
            new_parameters = {}
        for name, item in pars.items():
            fit_name = 'p' + str(name)
            if fit_name not in new_parameters.keys():
                new_parameters[fit_name] = item.raw_value

        return self._fit_function(x, **new_parameters, **kwargs)

    @abstractmethod
    def convert_to_pars_obj(self, par_list: Optional[Union[list]] = None):
        """
        Create an engine compatible container with the `Parameters` converted from the base object.

        :param par_list: If only a single/selection of parameter is required. Specify as a list
        :type par_list: List[str]
        :return: engine Parameters compatible object
        """

    @staticmethod
    @abstractmethod
    def convert_to_par_object(obj):
        """
        Convert an `easyCore.Objects.Base.Parameter` object to an engine Parameter object.
        """

    @abstractmethod
    def _set_parameter_fit_result(self, fit_result):
        """
        Update parameters to their final values and assign a std error to them.

        :param fit_result: Fit object which contains info on the fit
        :return: None
        :rtype: noneType
        """

    @abstractmethod
    def _gen_fit_results(self, fit_results, **kwargs) -> 'FitResults':
        """
        Convert fit results into the unified `FitResults` format.

        :param fit_result: Fit object which contains info on the fit
        :return: fit results container
        :rtype: FitResults
        """

    @abstractmethod
    def available_methods(self) -> List[str]:
        """
        Generate a list of available methods

        :return: List of available methods for minimization
        :rtype: List[str]
        """

    @staticmethod
    def _error_from_jacobian(jacobian: np.ndarray, residuals: np.ndarray, confidence: float = 0.95) -> np.ndarray:
        JtJi = np.linalg.inv(np.dot(jacobian.T, jacobian))
        # 1.96 is a 95% confidence value
        error_matrix = np.dot(JtJi, np.dot(jacobian.T,
                                  np.dot(np.diag(residuals ** 2), np.dot(jacobian, JtJi))))

        z = 1 - ((1 - confidence) / 2)
        z = stats.norm.pdf(z)
        error_matrix = z * np.sqrt(error_matrix)
        return error_matrix


class FitResults:
    """
    At the moment this is just a dummy way of unifying the returned fit parameters.
    """

    __slots__ = ['success', 'fitting_engine', 'fit_args',
                 'p', 'p0',
                 'x', 'x_matrices', 'y_obs', 'y_calc', 'residual',
                 'goodness_of_fit', 'engine_result']

    def __init__(self):
        self.success = False
        self.fitting_engine = None
        self.fit_args = {}
        self.p = {}
        self.p0 = {}
        self.x = np.ndarray([])
        self.x_matrices = np.ndarray([])
        self.y_obs = np.ndarray([])
        self.y_calc = np.ndarray([])
        self.goodness_of_fit = np.Inf
        self.residual = np.ndarray([])
        self.engine_result = None

    @property
    def n_pars(self):
        return len(self.p)

    @property
    def reduced_chi(self):
        return self.goodness_of_fit/(len(self.x) - self.n_pars)

    # def plot(self):


class NameConverter:

    def __init__(self):
        from easyCore import borg
        self._borg = borg

    def get_name_from_key(self, item_key: int) -> str:
        return getattr(self._borg.map.get_item_by_key(item_key), 'name', '')

    def get_item_from_key(self, item_key: int) -> object:
        return self._borg.map.get_item_by_key(item_key)

    def get_key(self, item: object) -> int:
        return self._borg.map.convert_id_to_key(item)


class FitError(Exception):

    def __init__(self, e: Exception = None):
        self.e = e

    def __str__(self) -> str:
        s = ''
        if self.e is not None:
            s = f'{self.e}\n'
        return s + 'Something has gone wrong with the fit'



#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

from easyCore.Objects.Graph import Graph
from easyCore.Utils.Hugger.Hugger import ScriptManager
from easyCore.Utils.classUtils import singleton
from easyCore.Utils.Logging import Logger


@singleton
class Borg:
    """
    Borg is the assimilated knowledge of `easyCore`. Every class based on `easyCore` gets brought
    into the collective.
    """
    __log = Logger()
    __map = Graph()
    __stack = None
    __debug = False

    def __init__(self):
        # Logger. This is so there's a unified logging interface
        self.log = self.__log
        # Debug. Global debugging level
        self.debug = self.__debug
        # Stack. This is where the undo/redo operations are stored.
        self.stack = self.__stack
        #
        self.script = ScriptManager()
        # Map. This is the conduit database between all borg species
        self.map = self.__map

    def instantiate_stack(self):
        """
        The undo/redo stack references the collective. Hence it has to be imported
        after initialization.

        :return: None
        :rtype: noneType
        """
        from easyCore.Utils.UndoRedo import UndoStack
        self.stack = UndoStack()



__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'


#  SPDX-FileCopyrightText: 2022 easyCore contributors  <core@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2021-2022 Contributors to the easyCore project <https://github.com/easyScience/easyCore>
"""
This module provides utility classes for string operations.
"""
# import re
from fractions import Fraction


def transformation_to_string(matrix, translation_vec=(0, 0, 0), components=('x', 'y', 'z'), c='', delim=','):
    """
    Convenience method. Given matrix returns string, e.g. x+2y+1/4
    :param matrix
    :param translation_vec
    :param components: either ('x', 'y', 'z') or ('a', 'b', 'c')
    :param c: optional additional character to print (used for magmoms)
    :param delim: delimiter
    :return: xyz string
    """
    parts = []
    for i in range(3):
        s = ''
        m = matrix[i]
        t = translation_vec[i]
        for j, dim in enumerate(components):
            if m[j] != 0:
                f = Fraction(m[j]).limit_denominator()
                if s != '' and f >= 0:
                    s += '+'
                if abs(f.numerator) != 1:
                    s += str(f.numerator)
                elif f < 0:
                    s += '-'
                s += c + dim
                if f.denominator != 1:
                    s += '/' + str(f.denominator)
        if t != 0:
            s += ('+' if (t > 0 and s != '') else '') + str(Fraction(t).limit_denominator())
        if s == '':
            s += '0'
        parts.append(s)
    return delim.join(parts)
